#     Copyright 2022. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import json
from queue import Queue
from random import choice
from re import fullmatch
from string import ascii_lowercase
from threading import Thread
from time import time, sleep
import ssl
import os

from simplejson import JSONDecodeError, dumps
import requests
from requests.auth import HTTPBasicAuth as HTTPBasicAuthRequest
from requests.exceptions import RequestException

from thingsboard_gateway.tb_utility.tb_loader import TBModuleLoader
from thingsboard_gateway.tb_utility.tb_utility import TBUtility
from thingsboard_gateway.connectors.connector import Connector
from thingsboard_gateway.tb_utility.tb_logger import init_logger

try:
    from requests import Timeout, request as regular_request
except ImportError:
    print("Requests library not found - installing...")
    TBUtility.install_package("requests")
    from requests import Timeout, request as regular_request

try:
    from aiohttp import web, BasicAuth
except ImportError:
    print('AIOHTTP library not found - installing...')
    TBUtility.install_package('aiohttp')
    from aiohttp import web, BasicAuth

requests.packages.urllib3.util.ssl_.DEFAULT_CIPHERS += ':ADH-AES128-SHA256'


class RESTConnector(Connector, Thread):
    def __init__(self, gateway, config, connector_type):
        super().__init__()
        self._default_converters = {
            "uplink": "JsonRESTUplinkConverter",
            "downlink": "JsonRESTDownlinkConverter"
        }
        self._events_type = {
            'STATISTICS_MESSAGE_RECEIVED': self.statistic_message_received,
            'STATISTICS_MESSAGE_SEND': self.statistic_message_send
        }
        self.__config = config
        self._connector_type = connector_type
        self.statistics = {'MessagesReceived': 0,
                           'MessagesSent': 0}
        self.setName(config.get("name", 'REST Connector ' + ''.join(choice(ascii_lowercase) for _ in range(5))))
        self.__gateway = gateway
        self.__log = init_logger(self.__gateway, self.name, self.__config.get('logLevel', 'INFO'))
        self._default_downlink_converter = TBModuleLoader.import_module(self._connector_type,
                                                                        self._default_converters['downlink'])
        self._default_uplink_converter = TBModuleLoader.import_module(self._connector_type,
                                                                      self._default_converters['uplink'])
        self.__USER_DATA = {}
        self._app = None
        self._connected = False
        self.__stopped = False
        self.daemon = True
        self.__attribute_type = {}
        self.__rpc_requests = []
        self.__attribute_updates = []
        self.__fill_requests_from_TB()

    def load_endpoints(self):
        endpoints = {}
        for mapping in self.__config.get("mapping"):
            converter = TBModuleLoader.import_module(self._connector_type,
                                                     mapping['converter'].get("extension",
                                                                              self._default_converters["uplink"]))
            endpoints.update({mapping['endpoint']: {"config": mapping, "converter": converter}})

        # configuring Attribute Request endpoints
        if len(self.__config.get('attributeRequests', [])):
            self.__attribute_type = {
                'client': self.__gateway.tb_client.client.gw_request_client_attributes,
                'shared': self.__gateway.tb_client.client.gw_request_shared_attributes
            }

            for attr in self.__config['attributeRequests']:
                config = {
                    'type': 'attributeRequest',
                    'function': self.__attribute_type[attr['type']],
                    'config': attr,
                }
                endpoints.update({attr['endpoint']: config})

        return endpoints

    def load_handlers(self):
        data_handlers = {
            "basic": BasicDataHandler,
            "anonymous": AnonymousDataHandler,
        }
        handlers = []
        mappings = self.__config.get("mapping", []) + self.__config.get('attributeRequests', [])
        for mapping in mappings:
            try:
                security_type = "anonymous" if mapping.get("security") is None else mapping["security"]["type"].lower()
                if security_type != "anonymous":
                    Users.add_user(mapping['endpoint'],
                                   mapping['security']['username'],
                                   mapping['security']['password'])
                for http_method in mapping['HTTPMethods']:
                    handler = data_handlers[security_type](self.collect_statistic_and_send, self.get_name(),
                                                           self.endpoints[mapping["endpoint"]], self.__log,
                                                           provider=self.__event_provider)
                    handlers.append(web.route(http_method, mapping['endpoint'], handler))
            except Exception as e:
                self.__log.error("Error on creating handlers - %s", str(e))
        self._app.add_routes(handlers)

    def open(self):
        self.__stopped = False
        self.start()

    def __run_server(self):
        self.endpoints = self.load_endpoints()

        self._app = web.Application(debug=self.__config.get('debugMode', False))

        ssl_context = None
        cert = None
        key = None
        if self.__config.get('SSL', False):
            if not self.__config.get('security'):
                if not os.path.exists('domain_srv.crt'):
                    from thingsboard_gateway.connectors.rest.ssl_generator import SSLGenerator
                    n = SSLGenerator(self.__config['host'])
                    n.generate_certificate()

                cert = 'domain_srv.crt'
                key = 'domain_srv.key'
            else:
                try:
                    cert = self.__config['security']['cert']
                    key = self.__config['security']['key']
                except KeyError as e:
                    self.__log.error('Provide certificate and key path!\n %s', e)

            ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            ssl_context.load_cert_chain(cert, key)

        self.load_handlers()
        web.run_app(self._app, host=self.__config['host'], port=self.__config['port'], handle_signals=False,
                    ssl_context=ssl_context, reuse_port=self.__config['port'], reuse_address=self.__config['host'])

    def run(self):
        self._connected = True

        try:
            self.__run_server()
        except Exception as e:
            self.__log.exception(e)

    def close(self):
        self.__stopped = True
        self._connected = False
        self.__log.reset()

    def get_name(self):
        return self.name

    def get_type(self):
        return self._connector_type

    def is_connected(self):
        return self._connected

    def get_config(self):
        return self.__config

    def on_attributes_update(self, content):
        try:
            for attribute_request in self.__attribute_updates:
                if fullmatch(attribute_request["deviceNameFilter"], content["device"]) and \
                        fullmatch(attribute_request["attributeFilter"], list(content["data"].keys())[0]):
                    converted_data = attribute_request["downlink_converter"].convert(attribute_request, content)
                    response_queue = Queue(1)
                    request_dict = {"config": {**attribute_request,
                                               **converted_data},
                                    "request": regular_request}
                    attribute_update_request_thread = Thread(target=self.__send_request,
                                                             args=(request_dict, response_queue, self.__log),
                                                             daemon=True,
                                                             name="Attribute request to %s" % (
                                                                 converted_data["url"]))
                    attribute_update_request_thread.start()
                    if not response_queue.empty():
                        response = response_queue.get_nowait()
                        self.__log.debug(response)
                    del response_queue

            # ONLY if initialized "response" section for endpoint
            # check if attribute update relates to some responseAttribute
            for endpoint in self.__config['mapping']:
                response_attribute = endpoint.get('response', {}).get('responseAttribute')
                if list(content['data'].keys())[0] == response_attribute:
                    BaseDataHandler.responses_queue.put(content['data'][response_attribute])
        except Exception as e:
            self.__log.exception(e)

    def server_side_rpc_handler(self, content):
        try:
            if content.get('data') is None:
                content['data'] = {'params': content['params'], 'method': content['method']}

            rpc_method = content['data']['method']

            # check if RPC type is connector RPC (can be only 'get' or 'set')
            try:
                (connector_type, rpc_method_name) = rpc_method.split('_')
                if connector_type == self._connector_type:
                    rpc_method = rpc_method_name
                    content['device'] = content['params'].split(' ')[0].split('=')[-1]
            except (IndexError, ValueError):
                pass

            # check if RPC method is reserved get/set
            if rpc_method == 'get' or rpc_method == 'set':
                device = content.get('device')
                params = {}
                for param in content['data']['params'].split(';'):
                    try:
                        (key, value) = param.split('=')
                    except ValueError:
                        continue

                    if key and value:
                        params[key] = value

                uplink_converter = self._default_uplink_converter
                downlink_converter = self._default_downlink_converter
                converted_data = downlink_converter.convert(params, content)

                request_dict = {'config': {**params, **converted_data}, 'request': regular_request,
                                'converter': uplink_converter}
                response = self.__send_request(request_dict, Queue(1), self.__log, with_queue=False)

                self.__log.debug('Response from RPC request: %s', response)
                self.__gateway.send_rpc_reply(device=device,
                                              req_id=content["data"].get('id'),
                                              content=response[2] if response and len(
                                                  response) >= 3 else response)
            else:
                for rpc_request in self.__rpc_requests:
                    if fullmatch(rpc_request["deviceNameFilter"], content["device"]) and \
                            fullmatch(rpc_request["methodFilter"], rpc_method):
                        converted_data = rpc_request["downlink_converter"].convert(rpc_request, content)

                        request_dict = {"config": {**rpc_request,
                                                   **converted_data},
                                        "request": regular_request}
                        request_dict["converter"] = request_dict["config"].get("uplink_converter")

                        response = self.__send_request(request_dict, Queue(1), self.__log, with_queue=False)

                        self.__log.debug('Response from RPC request: %s', response)
                        self.__gateway.send_rpc_reply(device=content["device"],
                                                      req_id=content["data"]["id"],
                                                      content=response[2] if response and len(
                                                          response) >= 3 else response)
        except Exception as e:
            self.__log.exception(e)

    def __event_provider(self, event_type):
        try:
            self._events_type[event_type]()
        except KeyError:
            self.__log.error('Unresolved event type')

    def statistic_message_received(self):
        self.statistics["MessagesReceived"] = self.statistics["MessagesReceived"] + 1

    def statistic_message_send(self):
        self.statistics["MessagesSent"] = self.statistics["MessagesSent"] + 1

    def collect_statistic_and_send(self, connector_name, data):
        self.statistics["MessagesReceived"] = self.statistics["MessagesReceived"] + 1
        self.__gateway.send_to_storage(connector_name, data)
        self.statistics["MessagesSent"] = self.statistics["MessagesSent"] + 1

    def __fill_requests_from_TB(self):
        requests_from_tb = {
            "attributeUpdates": self.__attribute_updates,
            "serverSideRpc": self.__rpc_requests,
        }
        for request_section in requests_from_tb:
            for request_config_object in self.__config.get(request_section, []):
                uplink_converter = TBModuleLoader.import_module(self._connector_type,
                                                                request_config_object.get("extension",
                                                                                          self._default_converters[
                                                                                              "uplink"]))(
                    request_config_object, self.__log)
                downlink_converter = TBModuleLoader.import_module(self._connector_type,
                                                                  request_config_object.get("extension",
                                                                                            self._default_converters[
                                                                                                "downlink"]))(
                    request_config_object, self.__log)
                request_dict = {**request_config_object,
                                "uplink_converter": uplink_converter,
                                "downlink_converter": downlink_converter,
                                }
                requests_from_tb[request_section].append(request_dict)

    def __send_request(self, request_dict, converter_queue, logger, with_queue=True):
        url = ""
        try:
            request_dict["next_time"] = time() + request_dict["config"].get("scanPeriod", 10)

            if str(request_dict["config"]["url"]).lower().startswith("http"):
                url = request_dict["config"]["url"]
            else:
                url = "http://" + request_dict["config"]["url"]

            logger.debug(url)
            security = None

            if request_dict["config"]["security"]["type"].lower() == "basic":
                security = HTTPBasicAuthRequest(request_dict["config"]["security"]["username"],
                                                request_dict["config"]["security"]["password"])

            request_timeout = request_dict["config"].get("timeout")

            data = {"data": request_dict["config"]["data"]}
            params = {
                "method": request_dict["config"].get("HTTPMethod", "GET"),
                "url": url,
                "timeout": request_timeout,
                "allow_redirects": request_dict["config"].get("allowRedirects", False),
                "verify": request_dict["config"].get("SSLVerify"),
                "auth": security,
                **data,
            }
            logger.debug(url)

            if request_dict["config"].get("httpHeaders") is not None:
                params["headers"] = request_dict["config"]["httpHeaders"]

            logger.debug("Request to %s will be sent", url)
            response = request_dict["request"](**params)
            data_to_storage = [url, request_dict["config"]["uplink_converter"]]

            if response and response.ok:
                try:
                    data_to_storage.append(response.json())
                except UnicodeDecodeError:
                    data_to_storage.append(response.content)
                except JSONDecodeError:
                    data_to_storage.append(response.content)

                if len(data_to_storage) == 3 and with_queue and not converter_queue.full():
                    converter_queue.put(data_to_storage)
                    self.statistics["MessagesReceived"] = self.statistics["MessagesReceived"] + 1
            else:
                logger.error("Request to URL: %s finished with code: %i. Cat information: http://http.cat/%i",
                             url,
                             response.status_code,
                             response.status_code)
                logger.debug("Response: %r", response.text)
                data_to_storage.append({"error": response.reason, "code": response.status_code})

                if with_queue:
                    converter_queue.put(data_to_storage)

                self.statistics["MessagesReceived"] = self.statistics["MessagesReceived"] + 1

            if not with_queue:
                return data_to_storage

        except Timeout:
            logger.error("Timeout error on request %s.", url)
        except RequestException as e:
            logger.error("Cannot connect to %s. Connection error.", url)
            logger.debug(e)
        except ConnectionError:
            logger.error("Cannot connect to %s. Connection error.", url)
        except Exception as e:
            logger.exception(e)


class BaseDataHandler:
    responses_queue = Queue()
    response_attribute_request = Queue()

    def __init__(self, send_to_storage, name, endpoint, logger, provider=None):
        self.log = logger
        self.send_to_storage = send_to_storage
        self.__name = name
        self.__endpoint = endpoint
        self.__provider = provider

        self.success_response = self.__endpoint['config'].get('response', {}).get('successResponse')
        self.unsuccessful_response = self.__endpoint['config'].get('response', {}).get('unsuccessfulResponse')
        self.response_expected = self.__endpoint['config'].get('response', {}).get('responseExpected', False)

    @property
    def name(self):
        return self.__name

    @property
    def endpoint(self):
        return self.__endpoint

    @staticmethod
    async def _convert_data_from_request(request):
        if request.method == 'GET':
            params = request.query

            return dict(params)
        else:
            try:
                json_data = await request.json()
            except json.decoder.JSONDecodeError:
                data = await request.post()
                if len(data):
                    json_data = dict(data)
                else:
                    json_data = await request.text()

            return json_data

    @staticmethod
    def modify_data_for_remote_response(data, modify):
        if modify:
            data['attributes'].append({'responseExpected': True})

    def get_response(self):
        if self.response_expected:
            time_point = time()
            while not time() - time_point >= self.endpoint['config'].get('response', {}).get('timeout', 120):
                if not BaseDataHandler.responses_queue.empty():
                    response = BaseDataHandler.responses_queue.get()
                    return web.Response(body=str(response), status=200)

                sleep(.2)

            return web.Response(body=str(self.unsuccessful_response) if self.unsuccessful_response else None,
                                status=408)

        return web.Response(body=str(self.success_response) if self.success_response else None, status=200)

    def process_attribute_request(self, data):
        if self.processed_attribute_request(data):
            time_point = time()
            while not time() - time_point >= self.endpoint['config']['timeout']:
                if not BaseDataHandler.response_attribute_request.empty():
                    self.__provider('STATISTICS_MESSAGE_SEND')
                    return web.Response(body=BaseDataHandler.response_attribute_request.get())

                sleep(.2)

            return web.Response(status=408)

    @staticmethod
    def attribute_request_callback(content, _):
        BaseDataHandler.response_attribute_request.put(dumps(content))

    def processed_attribute_request(self, data):
        if self.__endpoint.get('type') == 'attributeRequest':
            device_name_tags = TBUtility.get_values(self.__endpoint['config'].get("deviceNameExpression"), data,
                                                    get_tag=True)
            device_name_values = TBUtility.get_values(self.__endpoint['config'].get("deviceNameExpression"), data,
                                                      expression_instead_none=True)

            device_name = self.__endpoint['config'].get("deviceNameExpression")
            for (device_name_tag, device_name_value) in zip(device_name_tags, device_name_values):
                is_valid_key = "${" in self.__endpoint['config'].get("deviceNameExpression") and "}" in \
                               self.__endpoint['config'].get("deviceNameExpression")
                device_name = device_name.replace('${' + str(device_name_tag) + '}',
                                                  str(device_name_value)) \
                    if is_valid_key else device_name_tag

            if not device_name or device_name == '':
                return False

            found_attribute_names = list(filter(lambda x: x is not None,
                                                TBUtility.get_values(
                                                    self.__endpoint['config'].get("attributeNameExpression"),
                                                    data)))

            if found_attribute_names is None:
                return False

            self.__endpoint['function'](device_name, found_attribute_names, self.attribute_request_callback)
            self.__provider('STATISTICS_MESSAGE_RECEIVED')
            return True

        return False


class AnonymousDataHandler(BaseDataHandler):
    async def __call__(self, request: web.Request):
        json_data = await self._convert_data_from_request(request)

        if not json_data and not len(request.query):
            return web.Response(body=str(self.unsuccessful_response) if self.unsuccessful_response else None,
                                status=415)

        endpoint_config = self.endpoint['config']
        if request.method.upper() not in [method.upper() for method in endpoint_config['HTTPMethods']]:
            return web.Response(body=str(self.unsuccessful_response) if self.unsuccessful_response else None,
                                status=405)

        json_data.update(dict(request.query))
        data = json_data

        # check if request is Attribute Request type
        result = self.process_attribute_request(data)
        if isinstance(result, web.Response):
            return result

        try:
            self.log.info("CONVERTER CONFIG: %r", endpoint_config['converter'])
            converter = self.endpoint['converter'](endpoint_config['converter'], self.log)
            converted_data = converter.convert(config=endpoint_config['converter'], data=data)

            self.modify_data_for_remote_response(converted_data, self.response_expected)

            self.send_to_storage(self.name, converted_data)
            self.log.info("CONVERTED_DATA: %r", converted_data)

            return self.get_response()
        except Exception as e:
            self.log.exception("Error while post to anonymous handler: %s", e)
            return web.Response(body=str(self.success_response) if self.success_response else None, status=500)


class BasicDataHandler(BaseDataHandler):
    def verify(self, username, password):
        if not username and password:
            return False
        return Users.validate_user_credentials(self.endpoint['config']['endpoint'], username, password)

    async def __call__(self, request: web.Request):
        if request.headers.get('Authorization') is None:
            return web.Response(body=str(self.unsuccessful_response) if self.unsuccessful_response else None,
                                status=401)

        auth = BasicAuth.decode(request.headers['Authorization'])
        if self.verify(auth.login, auth.password):
            json_data = await self._convert_data_from_request(request)

            if not json_data:
                return web.Response(body=str(self.unsuccessful_response) if self.unsuccessful_response else None,
                                    status=415)

            endpoint_config = self.endpoint['config']

            if request.method.upper() not in [method.upper() for method in endpoint_config['HTTPMethods']]:
                return web.Response(body=str(self.unsuccessful_response) if self.unsuccessful_response else None,
                                    status=405)

            json_data.update(dict(request.query))
            data = json_data

            # check if request is Attribute Request type
            result = self.process_attribute_request(data)
            if isinstance(result, web.Response):
                return result

            try:
                self.log.info("CONVERTER CONFIG: %r", endpoint_config['converter'])
                converter = self.endpoint['converter'](endpoint_config['converter'], self.log)
                converted_data = converter.convert(config=endpoint_config['converter'], data=data)

                self.modify_data_for_remote_response(converted_data, self.response_expected)

                self.send_to_storage(self.name, converted_data)
                self.log.info("CONVERTED_DATA: %r", converted_data)

                return self.get_response()
            except Exception as e:
                self.log.exception("Error while post to basic handler: %s", e)
                return web.Response(body=str(self.unsuccessful_response) if self.unsuccessful_response else None,
                                    status=500)

        return web.Response(body=str(self.unsuccessful_response) if self.unsuccessful_response else None, status=401)


class Users:
    USER_DATA = {}

    @classmethod
    def add_user(cls, endpoint, username, password):
        cls.USER_DATA.update({endpoint: {username: password}})

    @classmethod
    def validate_user_credentials(cls, endpoint, username, password):
        result = False
        if cls.USER_DATA.get(endpoint) is not None and cls.USER_DATA[endpoint].get(username) == password:
            result = True
        return result



import os
import inspect
import re
import subprocess
import sys
from time import time, sleep


class HotReloader:
    def __init__(self, function):
        self._stopped = False
        self._file_pattern = r'^(?!.*.(pyc|log|\d)$).*$'
        self._exclude_files = [
            'connected_devices.json',
            'persistent_keys.json'
        ]
        self._runnable_function = function
        self._poll_interval = 1
        self._process = None

        self._runnable_function_path = os.path.abspath(inspect.getfile(self._runnable_function))
        self._root_path = '/'.join(os.path.abspath(inspect.getfile(self._runnable_function)).split('/')[:-2])
        self._files = self._find_files()

        self._last_poll_time = 0

        self.run()

    def _find_files(self):
        files = []
        for root, _, filenames in os.walk(self._root_path):
            for filename in filenames:
                if re.match(self._file_pattern, filename) and filename not in self._exclude_files:
                    file_path = os.path.join(root, filename)
                    files.append((file_path, os.stat(file_path).st_mtime))

        return files

    def run(self):
        try:
            self.execute()

            while not self._stopped:
                if time() - self._last_poll_time >= self._poll_interval:
                    files = self._find_files()
                    if files != self._files:
                        print('Reloading....')
                        self.execute()
                        self._files = files

                    self._last_poll_time = time()

                sleep(.2)
        except KeyboardInterrupt:
            self._stopped = True

    def execute(self):
        if self._process is not None:
            self._process.kill()
            self._process.wait()

        self._process = subprocess.Popen([sys.executable, self._runnable_function_path])



#     Copyright 2022. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from pprint import pformat
from re import findall

from thingsboard_gateway.connectors.ble.ble_uplink_converter import BLEUplinkConverter
from thingsboard_gateway.gateway.statistics_service import StatisticsService


class BytesBLEUplinkConverter(BLEUplinkConverter):
    def __init__(self, config, logger):
        self._log = logger
        self.__config = config

    @StatisticsService.CollectStatistics(start_stat_type='receivedBytesFromDevices',
                                         end_stat_type='convertedBytesFromDevice')
    def convert(self, config, data):
        if data is None:
            return {}

        dict_result = {
            "deviceName": self.__config['deviceName'],
            "deviceType": self.__config['deviceType']
        }

        try:
            dict_result["telemetry"] = []
            dict_result["attributes"] = []

            for section in ('telemetry', 'attributes'):
                for item in data[section]:
                    try:
                        expression_arr = findall(r'\[[^\s][0-9:]*]', item['valueExpression'])
                        converted_data = item['valueExpression']

                        for exp in expression_arr:
                            indexes = exp[1:-1].split(':')

                            data_to_replace = ''
                            if len(indexes) == 2:
                                from_index, to_index = indexes
                                concat_arr = item['data'][
                                             int(from_index) if from_index != '' else None:int(
                                                 to_index) if to_index != '' else None]
                                for sub_item in concat_arr:
                                    data_to_replace += str(sub_item)
                            else:
                                data_to_replace += str(item['data'][int(indexes[0])])

                            converted_data = converted_data.replace(exp, data_to_replace)

                        if item.get('key') is not None:
                            dict_result[section].append({item['key']: converted_data})
                        else:
                            self._log.error('Key for %s not found in config: %s', config['type'], config[section])
                    except Exception as e:
                        self._log.exception('\nException caught when processing data for %s\n\n%s', pformat(config), e)

        except Exception as e:
            self._log.exception(e)

        self._log.debug(dict_result)
        return dict_result



#     Copyright 2022. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from time import time
from datetime import timezone

from thingsboard_gateway.connectors.opcua_asyncio.opcua_converter import OpcUaConverter
from asyncua.ua.uatypes import LocalizedText, VariantType

DATA_TYPES = {
    'attributes': 'attributes',
    'timeseries': 'telemetry'
}


class OpcUaUplinkConverter(OpcUaConverter):
    def __init__(self, config, logger):
        self._log = logger
        self.__config = config
        self.data = {
            'deviceName': self.__config['device_name'],
            'deviceType': 'default',
            'attributes': [],
            'telemetry': [],
        }

    def clear_data(self):
        self.data = {
            'deviceName': self.__config['device_name'],
            'deviceType': 'default',
            'attributes': [],
            'telemetry': [],
        }

    def get_data(self):
        if len(self.data['attributes']) or len(self.data['telemetry']):
            data_list = []
            device_names = self.__config.get('device_names')
            if device_names:
                for device in device_names:
                    self.data['deviceName'] = device
                    data_list.append(self.data)

                return data_list

            return [self.data]

        return None

    def convert(self, config, val):
        if not val:
            return

        data = val.Value.Value

        if data is not None:
            if isinstance(data, LocalizedText):
                data = data.Text
            elif val.Value.VariantType == VariantType.ExtensionObject:
                data = str(data)
            elif val.Value.VariantType == VariantType.DateTime:
                if data.tzinfo is None:
                    data = data.replace(tzinfo=timezone.utc)
                data = data.isoformat()

            if config['section'] == 'timeseries':
                if val.SourceTimestamp:
                    timestamp = int(val.SourceTimestamp.replace(tzinfo=timezone.utc).timestamp() * 1000)
                elif val.ServerTimestamp:
                    timestamp = int(val.ServerTimestamp.replace(tzinfo=timezone.utc).timestamp() * 1000)
                else:
                    timestamp = int(time() * 1000)

                self.data[DATA_TYPES[config['section']]].append({'ts': timestamp, 'values': {config['key']: data}})
            else:
                self.data[DATA_TYPES[config['section']]].append({config['key']: data})

            self._log.debug('Converted data: %s', self.data)



#     Copyright 2023. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import logging
from time import sleep
from threading import Thread


def init_logger(gateway, name, level):
    """
    For creating a Logger with all config automatically
    Create a Logger manually only if you know what you are doing!
    """
    log = TbLogger(name=name, gateway=gateway)

    if hasattr(gateway, 'remote_handler'):
        log.addHandler(gateway.remote_handler)
        log.setLevel(gateway.main_handler.level)
        gateway.remote_handler.add_logger(name)

    if hasattr(gateway, 'main_handler'):
        log.addHandler(gateway.main_handler)
        log.setLevel(gateway.remote_handler.level)

    log_level_conf = level
    if log_level_conf:
        log_level = logging.getLevelName(log_level_conf)
        log.setLevel(log_level)

    return log


class TbLogger(logging.Logger):
    ALL_ERRORS_COUNT = 0
    IS_ALL_ERRORS_COUNT_RESET = False

    def __init__(self, name, gateway=None, level=logging.NOTSET):
        super(TbLogger, self).__init__(name=name, level=level)
        self.propagate = True
        self.parent = self.root
        self._gateway = gateway
        self.errors = 0
        self.attr_name = self.name + '_ERRORS_COUNT'
        self._is_on_init_state = True
        self._errors_sender_thread = Thread(name='Log Errors Sender', daemon=True, target=self._send_errors)
        self._errors_sender_thread.start()

    def reset(self):
        """
        !!!Need to be called manually in the connector 'close' method!!!
        """
        TbLogger.ALL_ERRORS_COUNT = TbLogger.ALL_ERRORS_COUNT - self.errors

    @property
    def gateway(self):
        return self._gateway

    @gateway.setter
    def gateway(self, gateway):
        self._gateway = gateway

    def _send_errors(self):
        is_tb_client = False

        while not self._gateway:
            sleep(1)

        while not is_tb_client:
            is_tb_client = hasattr(self._gateway, 'tb_client')
            sleep(1)

        if not TbLogger.IS_ALL_ERRORS_COUNT_RESET and self._gateway.tb_client.is_connected():
            self._gateway.tb_client.client.send_telemetry(
                {self.attr_name: 0, 'ALL_ERRORS_COUNT': 0}, quality_of_service=0)
            TbLogger.IS_ALL_ERRORS_COUNT_RESET = True
        self._is_on_init_state = False

    def error(self, msg, *args, **kwargs):
        kwargs['stacklevel'] = 2
        super(TbLogger, self).error(msg, *args, **kwargs)
        self._send_error_count()

    def exception(self, msg, *args, **kwargs) -> None:
        attr_name = kwargs.pop('attr_name', None)
        kwargs['stacklevel'] = 2
        super(TbLogger, self).exception(msg, *args, **kwargs)
        self._send_error_count(error_attr_name=attr_name)

    def _send_error_count(self, error_attr_name=None):
        TbLogger.ALL_ERRORS_COUNT += 1
        self.errors += 1

        while self._is_on_init_state:
            sleep(.2)

        if self._gateway and hasattr(self._gateway, 'tb_client'):
            if error_attr_name:
                error_attr_name = error_attr_name + '_ERRORS_COUNT'
            else:
                error_attr_name = self.attr_name

            self._gateway.tb_client.client.send_telemetry(
                {error_attr_name: self.errors, 'ALL_ERRORS_COUNT': TbLogger.ALL_ERRORS_COUNT})



#     Copyright 2022. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import asyncio
import base64
import re
import ssl
from queue import Queue
from threading import Thread
from random import choice
from string import ascii_lowercase
from time import sleep

from simplejson import dumps

from thingsboard_gateway.connectors.connector import Connector
from thingsboard_gateway.gateway.statistics_service import StatisticsService
from thingsboard_gateway.tb_utility.tb_utility import TBUtility
from thingsboard_gateway.tb_utility.tb_logger import init_logger

try:
    import ocpp
except ImportError:
    print('OCPP library not found - installing...')
    TBUtility.install_package("ocpp")
    import ocpp

try:
    import websockets
except ImportError:
    print('websockets library not found - installing...')
    TBUtility.install_package("websockets")
    import websockets

from ocpp.v16 import call
from thingsboard_gateway.connectors.ocpp.charge_point import ChargePoint


class NotAuthorized(Exception):
    """Charge Point not authorized"""


class OcppConnector(Connector, Thread):
    DATA_TO_CONVERT = Queue(-1)
    DATA_TO_SEND = Queue(-1)

    def __init__(self, gateway, config, connector_type):
        super().__init__()
        self._config = config
        self._central_system_config = config['centralSystem']
        self._charge_points_config = config.get('chargePoints', [])
        self._connector_type = connector_type
        self.statistics = {'MessagesReceived': 0,
                           'MessagesSent': 0}
        self._gateway = gateway
        self.setName(self._config.get("name", 'OCPP Connector ' + ''.join(choice(ascii_lowercase) for _ in range(5))))
        self._log = init_logger(self._gateway, self.name, self._config.get('logLevel', 'INFO'))

        self._default_converters = {'uplink': 'OcppUplinkConverter'}
        self._server = None
        self._connected_charge_points = []

        self._ssl_context = None
        try:
            if self._central_system_config['connection']['type'].lower() == 'tls':
                self._ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
                pem_file = self._central_system_config['connection']['cert']
                key_file = self._central_system_config['connection']['key']
                password = self._central_system_config['connection'].get('password')
                self._ssl_context.load_cert_chain(pem_file, key_file, password=password)
        except Exception as e:
            self._log.exception(e)
            self._log.warning('TLS connection not set!')
            self._ssl_context = None

        self._data_convert_thread = Thread(name='Convert Data Thread', daemon=True, target=self._process_data)
        self._data_send_thread = Thread(name='Send Data Thread', daemon=True, target=self._send_data)

        self.__loop = asyncio.new_event_loop()

        self.__connected = False
        self.__stopped = False
        self.daemon = True

    def open(self):
        self.__stopped = False
        self.start()
        self._log.info("Starting OCPP Connector")

    def get_type(self):
        return self._connector_type

    def run(self):
        self._data_convert_thread.start()
        self._data_send_thread.start()

        self.__loop.run_until_complete(self.start_server())

    async def start_server(self):
        host = self._central_system_config.get('host', '0.0.0.0')
        port = self._central_system_config.get('port', 9000)
        self._server = await websockets.serve(self.on_connect, host, port, subprotocols=['ocpp1.6'],
                                              ssl=self._ssl_context)
        self.__connected = True
        self._log.info('Central System is running on %s:%d', host, port)

        await self._server.wait_closed()

    def _auth(self, websocket):
        for sec in self._central_system_config['security']:
            if sec['type'].lower() == 'token' and websocket.request_headers['authorization'] in sec['tokens']:
                self._log.debug('Got Authorization: %s', websocket.request_headers['authorization'])
                return
            elif sec['type'].lower() == 'basic':
                for cred in sec['credentials']:
                    token = 'Basic {0}'.format(
                        base64.b64encode(bytes(cred['username'] + ':' + cred['password'], 'utf-8')).decode('ascii'))
                    if websocket.request_headers['authorization'] == token:
                        self._log.debug('Got Authorization: %s', websocket.request_headers['authorization'])
                        return

        raise NotAuthorized('Charge Point not authorized')

    async def on_connect(self, websocket, path):
        """ For every new charge point that connects, create a ChargePoint instance
        and start listening for messages.

        """
        requested_protocols = None
        try:
            requested_protocols = websocket.request_headers[
                'Sec-WebSocket-Protocol']
        except KeyError:
            self._log.info("Client hasn't requested any Subprotocol. "
                           "Closing Connection")

        # Authorize Charge Point before accept connection
        if self._central_system_config.get('security'):
            try:
                self._auth(websocket)
            except NotAuthorized as e:
                self._log.error(e)
                return await websocket.close()

        if websocket.subprotocol:
            self._log.info("Protocols Matched: %s", websocket.subprotocol)
        else:
            # In the websockets lib if no subprotocols are supported by the
            # client and the server, it proceeds without a subprotocol,
            # so we have to manually close the connection.
            self._log.warning('Protocols Mismatched | Expected Subprotocols: %s,'
                              ' but client supports  %s | Closing connection',
                              websocket.available_subprotocols,
                              requested_protocols)
            return await websocket.close()

        cp_host = None
        cp_port = None
        try:
            (cp_host, cp_port) = websocket.remote_address
        except ValueError:
            pass

        charge_point_id = path.strip('/')

        # check if charge point can be connected
        (is_valid, cp_config) = await self._is_charge_point_valid(charge_point_id, host=cp_host, port=cp_port)
        if is_valid:
            uplink_converter_name = cp_config.get('extension', self._default_converters['uplink'])
            cp = ChargePoint(charge_point_id, websocket, {**cp_config, 'uplink_converter_name': uplink_converter_name},
                             OcppConnector._callback, self._log)
            cp.authorized = True

            self._log.info('Connected Charge Point with id: %s', charge_point_id)
            self._connected_charge_points.append(cp)

            try:
                await cp.start()
            except websockets.ConnectionClosed:
                self._connected_charge_points.pop(self._connected_charge_points.index(cp))

    async def _is_charge_point_valid(self, charge_point_id, **kwargs):
        for cp_config in self._charge_points_config:
            if re.match(cp_config['idRegexpPattern'], charge_point_id):
                host = cp_config.get('host')
                port = cp_config.get('port')
                if host or port:
                    if not re.match(host, kwargs.get('host')) or not re.match(port, kwargs.get('port')):
                        return False, None

                return True, cp_config

        return False, None

    def close(self):
        self.__stopped = True
        self.__connected = False

        task = self.__loop.create_task(self._close_cp_connections())
        while not task.done():
            sleep(.2)

        self._log.info('%s has been stopped.', self.get_name())
        self._log.reset()

    async def _close_cp_connections(self):
        for cp in self._connected_charge_points:
            await cp.close()

    def get_name(self):
        return self.name

    def is_connected(self):
        return self.__connected

    @classmethod
    def _callback(cls, data):
        cls.DATA_TO_CONVERT.put(data)

    def _process_data(self):
        while not self.__stopped:
            if not self.DATA_TO_CONVERT.empty():
                self.statistics['MessagesReceived'] += 1
                (converter, config, data) = self.DATA_TO_CONVERT.get()
                self._log.debug('Data from Charge Point: %s', data)
                converted_data = converter.convert(config, data)
                if converted_data and (converted_data.get('attributes') or converted_data.get('telemetry')):
                    self.DATA_TO_SEND.put(converted_data)

            sleep(.001)

    def _send_data(self):
        while not self.__stopped:
            if not self.DATA_TO_SEND.empty():
                converted_data = self.DATA_TO_SEND.get()
                self._gateway.send_to_storage(self.name, converted_data)
                self.statistics['MessagesSent'] += 1
                self._log.info("Data to ThingsBoard: %s", converted_data)

            sleep(.001)

    @staticmethod
    async def _send_request(cp, request):
        return await cp.call(request)

    @StatisticsService.CollectAllReceivedBytesStatistics(start_stat_type='allReceivedBytesFromTB')
    def on_attributes_update(self, content):
        self._log.debug('Got attribute update: %s', content)

        try:
            charge_point = tuple(filter(lambda cp: cp.name == content['device'], self._connected_charge_points))[0]
        except IndexError:
            self._log.error('Charge Point with name %s not found!', content['device'])
            return

        try:
            for attribute_update_config in charge_point.config.get('attributeUpdates', []):
                for (attr_key, attr_value) in content['data'].items():
                    if attr_key == attribute_update_config['attributeOnThingsBoard']:
                        data = attribute_update_config["valueExpression"] \
                            .replace("${attributeKey}", str(attr_key)) \
                            .replace("${attributeValue}", str(attr_value))
                        request = call.DataTransferPayload('1', data=data)

                        task = self.__loop.create_task(self._send_request(charge_point, request))
                        while not task.done():
                            sleep(.2)

                        self._log.debug(task.result())
        except Exception as e:
            self._log.exception(e)

    @StatisticsService.CollectAllReceivedBytesStatistics(start_stat_type='allReceivedBytesFromTB')
    def server_side_rpc_handler(self, content):
        self._log.debug('Got RPC: %s', content)

        try:
            charge_point = tuple(filter(lambda cp: cp.name == content['device'], self._connected_charge_points))[0]
        except IndexError:
            self._log.error('Charge Point with name %s not found!', content['device'])
            return

        try:
            for rpc in charge_point.config.get('serverSideRpc', []):
                if rpc['methodRPC'] == content['data']['method']:
                    data_to_send_tags = TBUtility.get_values(rpc.get('valueExpression'), content['data'],
                                                             'params',
                                                             get_tag=True)
                    data_to_send_values = TBUtility.get_values(rpc.get('valueExpression'), content['data'],
                                                               'params',
                                                               expression_instead_none=True)

                    data_to_send = rpc.get('valueExpression')
                    for (tag, value) in zip(data_to_send_tags, data_to_send_values):
                        data_to_send = data_to_send.replace('${' + tag + '}', dumps(value))

                    request = call.DataTransferPayload('1', data=data_to_send)

                    task = self.__loop.create_task(self._send_request(charge_point, request))
                    while not task.done():
                        sleep(.2)

                    if rpc.get('withResponse', True):
                        self._gateway.send_rpc_reply(content["device"], content["data"]["id"], str(task.result()))

                        return
        except Exception as e:
            self._log.exception(e)

    def get_config(self):
        return {'CS': self._central_system_config, 'CP': self._charge_points_config}



#     Copyright 2022. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import re


class Device:
    def __init__(self, path, name, config, converter, converter_for_sub, logger):
        self._log = logger
        self.path = path
        self.name = name
        self.config = config
        self.converter = converter
        self.converter_for_sub = converter_for_sub
        self.values = {
            'timeseries': [],
            'attributes': []
        }

        self.load_values()

    def __repr__(self):
        return f'{self.path}'

    def load_values(self):
        for section in ('attributes', 'timeseries'):
            for node_config in self.config.get(section, []):
                try:
                    if re.search(r"(ns=\d+;[isgb]=[^}]+)", node_config['path']):
                        child = re.search(r"(ns=\d+;[isgb]=[^}]+)", node_config['path'])
                        self.values[section].append({'path': child.groups()[0], 'key': node_config['key']})
                    elif re.search(r"\${([A-Za-z.:\\\d]+)}", node_config['path']):
                        child = re.search(r"\${([A-Za-z.:\\\d]+)", node_config['path'])
                        self.values[section].append(
                            {'path': self.path + child.groups()[0].split('\\.'), 'key': node_config['key']})

                except KeyError as e:
                    self._log.error('Invalid config for %s (key %s not found)', node_config, e)



#      Copyright 2022. ThingsBoard
#  #
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#  #
#          http://www.apache.org/licenses/LICENSE-2.0
#  #
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.
from time import time
from typing import Union

from simplejson import dumps

from thingsboard_gateway.connectors.converter import Converter, log
from thingsboard_gateway.gateway.constant_enums import DownlinkMessageType
from thingsboard_gateway.gateway.proto.messages_pb2 import *


class GrpcDownlinkConverter(Converter):
    def __init__(self):
        self.__conversion_methods = {
            DownlinkMessageType.Response: self.__convert_response_msg,
            DownlinkMessageType.ConnectorConfigurationMsg: self.__convert_connector_configuration_msg,
            DownlinkMessageType.GatewayAttributeUpdateNotificationMsg: self.__convert_gateway_attribute_update_notification_msg,
            DownlinkMessageType.GatewayAttributeResponseMsg: self.__convert_gateway_attribute_response_msg,
            DownlinkMessageType.GatewayDeviceRpcRequestMsg: self.__convert_gateway_device_rpc_request_msg,
            DownlinkMessageType.UnregisterConnectorMsg: self.__convert_unregister_connector_msg,
            DownlinkMessageType.ConnectorGetConnectedDevicesResponseMsg: self.__convert_get_connected_devices_msg
            }

    def convert(self, config, msg):
        try:
            basic_msg = FromServiceMessage()
            message_types = config.get("message_type")
            additional_message = config.get("additional_message")
            if not isinstance(message_types, list):
                message_types = [message_types]
            for message_type in message_types:
                self.__conversion_methods[message_type](basic_msg, msg, additional_message)
            return basic_msg
        except Exception as e:
            log.exception("[GRPC] ", e)
            return None

    @staticmethod
    def __convert_response_msg(basic_msg, msg, additional_message):
        if additional_message is not None:
            if additional_message.HasField('gatewayTelemetryMsg'):
                additional_message.gatewayTelemetryMsg.MergeFrom(GatewayTelemetryMsg())
            elif additional_message.HasField("gatewayAttributesMsg"):
                additional_message.gatewayAttributesMsg.MergeFrom(GatewayAttributesMsg())
            else:
                basic_msg.response.connectorMessage.MergeFrom(additional_message)
        basic_msg.response.status = ResponseStatus.Value(msg.name)

    @staticmethod
    def __convert_connector_configuration_msg(basic_msg, msg, additional_data=None):
        pass

    @staticmethod
    def __convert_gateway_attribute_update_notification_msg(basic_msg, msg, additional_data=None):
        ts = int(time()*1000)
        gw_attr_upd_notify_msg = GatewayAttributeUpdateNotificationMsg()
        gw_attr_upd_notify_msg.deviceName = msg['device']
        attr_notify_msg = AttributeUpdateNotificationMsg()
        for shared_attribute in msg['data']:
            if len(msg["data"]) == 1 and msg["data"].get("deleted") is not None and isinstance(msg["data"]["deleted"], list):
                attr_notify_msg.sharedDeleted.extend(msg["data"]["deleted"])
                break
            ts_kv_proto = TsKvProto()
            ts_kv_proto.ts = ts
            kv = GrpcDownlinkConverter.__get_key_value_proto_value(shared_attribute, msg['data'][shared_attribute])
            ts_kv_proto.kv.MergeFrom(kv)
            attr_notify_msg.sharedUpdated.extend([ts_kv_proto])
        gw_attr_upd_notify_msg.notificationMsg.MergeFrom(attr_notify_msg)
        basic_msg.gatewayAttributeUpdateNotificationMsg.MergeFrom(gw_attr_upd_notify_msg)
        return basic_msg

    @staticmethod
    def __convert_gateway_attribute_response_msg(basic_msg, msg, additional_data=None):
        attrs_resp_msg = GatewayAttributesResponseMsg()
        attrs_resp_msg.requestId = additional_data["request_id"]
        if additional_data.get("error") is not None:
            attrs_resp_msg.error = additional_data["error"]
        if additional_data.get("key") is not None:  # Single key requested
            if additional_data["client"]:
                attrs_resp_msg.clientAttributeList.extend([GrpcDownlinkConverter.__get_key_value_proto_value(additional_data["key"], msg.get("value"))])
            else:
                attrs_resp_msg.sharedAttributeList.extend([GrpcDownlinkConverter.__get_key_value_proto_value(additional_data["key"], msg.get("value"))])
        else:  # Several keys requested
            for key, value in msg["values"].items():
                if additional_data["client"]:
                    attrs_resp_msg.clientAttributeList.extend([GrpcDownlinkConverter.__get_key_value_proto_value(key, value)])
                else:
                    attrs_resp_msg.sharedAttributeList.extend([GrpcDownlinkConverter.__get_key_value_proto_value(key, value)])
        gw_attr_resp_msg = GatewayAttributeResponseMsg()
        gw_attr_resp_msg.deviceName = msg["device"]
        gw_attr_resp_msg.responseMsg.MergeFrom(attrs_resp_msg)
        basic_msg.gatewayAttributeResponseMsg.MergeFrom(gw_attr_resp_msg)
        return basic_msg


    @staticmethod
    def __convert_gateway_device_rpc_request_msg(basic_msg, msg, additional_data=None):
        msg_data = msg['data']
        gw_to_device_rpc = GatewayDeviceRpcRequestMsg()
        gw_to_device_rpc.deviceName = msg['device']
        rpc_request_msg = ToDeviceRpcRequestMsg()
        rpc_request_msg.requestId = msg_data['id']
        rpc_request_msg.methodName = msg_data['method']
        if isinstance(msg_data['params'], dict):
            rpc_request_msg.params = dumps(msg_data['params'])
        else:
            rpc_request_msg.params = str(msg_data['params'])
        gw_to_device_rpc.rpcRequestMsg.MergeFrom(rpc_request_msg)
        basic_msg.gatewayDeviceRpcRequestMsg.MergeFrom(gw_to_device_rpc)
        return basic_msg

    @staticmethod
    def __convert_unregister_connector_msg(basic_msg, msg, additional_data=None):
        if msg is None:
            msg = b''
        unreg_msg = UnregisterConnectorMsg()
        unreg_msg.connectorKey = msg
        basic_msg.unregisterConnectorMsg.MergeFrom(unreg_msg)
        return basic_msg

    @staticmethod
    def __convert_get_connected_devices_msg(basic_msg, msg, additional_data=None):
        status = ResponseStatus.Value("SUCCESS")
        if additional_data is None:
            status = ResponseStatus.Value("FAILURE")
            additional_data = {}
        connector_devices_response_msg = ConnectorGetConnectedDevicesResponseMsg()
        for device_name, device_type in additional_data.items():
            device_info_msg = ConnectorDeviceInfo()
            device_info_msg.deviceName = device_name
            device_info_msg.deviceType = device_type
            connector_devices_response_msg.connectorDevices.extend([device_info_msg])
        basic_msg.connectorGetConnectedDevicesResponseMsg.MergeFrom(connector_devices_response_msg)
        basic_msg.response.connectorMessage.connectorGetConnectedDevicesMsg.MergeFrom(ConnectorGetConnectedDevicesMsg())
        basic_msg.response.status = status
        return basic_msg

    @staticmethod
    def __get_key_value_proto_value(key: str, value: Union[str, bool, int, float, dict]) -> KeyValueProto:
        key_value_proto = KeyValueProto()
        key_value_proto.key = key
        if isinstance(value, bool):
            key_value_proto.type = KeyValueType.BOOLEAN
            key_value_proto.bool_v = value
        elif isinstance(value, int):
            key_value_proto.type = KeyValueType.LONG_V
            key_value_proto.long_v = value
        elif isinstance(value, float):
            key_value_proto.type = KeyValueType.DOUBLE_V
            key_value_proto.double_v = value
        elif isinstance(value, str):
            key_value_proto.type = KeyValueType.STRING_V
            key_value_proto.string_v = value
        elif isinstance(value, dict):
            key_value_proto.type = KeyValueType.JSON_V
            key_value_proto.json_v = dumps(value)
        return key_value_proto



#     Copyright 2022. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import json
from re import findall
from time import time

from thingsboard_gateway.connectors.xmpp.xmpp_converter import XmppConverter
from thingsboard_gateway.tb_utility.tb_utility import TBUtility
from thingsboard_gateway.gateway.statistics_service import StatisticsService


class XmppUplinkConverter(XmppConverter):
    def __init__(self, config, logger):
        self._log = logger
        self.__config = config
        self._datatypes = {"attributes": "attributes",
                           "timeseries": "telemetry"}

    def _convert_json(self, val):
        dict_result = {}
        try:
            data = json.loads(val)

            device_name_tags = TBUtility.get_values(self.__config.get("deviceNameExpression"), data,
                                                    get_tag=True)
            device_name_values = TBUtility.get_values(self.__config.get("deviceNameExpression"), data,
                                                      expression_instead_none=True)

            dict_result['deviceName'] = self.__config.get("deviceNameExpression")
            for (device_name_tag, device_name_value) in zip(device_name_tags, device_name_values):
                is_valid_key = "${" in self.__config.get("deviceNameExpression") and "}" in \
                               self.__config.get("deviceNameExpression")
                dict_result['deviceName'] = dict_result['deviceName'].replace('${' + str(device_name_tag) + '}',
                                                                              str(device_name_value)) \
                    if is_valid_key else device_name_tag

            device_type_tags = TBUtility.get_values(self.__config.get("deviceTypeExpression"), data,
                                                    get_tag=True)
            device_type_values = TBUtility.get_values(self.__config.get("deviceTypeExpression"), data,
                                                      expression_instead_none=True)

            dict_result["deviceType"] = self.__config.get("deviceTypeExpression")
            for (device_type_tag, device_type_value) in zip(device_type_tags, device_type_values):
                is_valid_key = "${" in self.__config.get("deviceTypeExpression") and "}" in \
                               self.__config.get("deviceTypeExpression")
                dict_result["deviceType"] = dict_result["deviceType"].replace('${' + str(device_type_tag) + '}',
                                                                              str(device_type_value)) \
                    if is_valid_key else device_type_tag

            for datatype in self._datatypes:
                dict_result[self._datatypes[datatype]] = []
                for datatype_config in self.__config.get(datatype, []):
                    values = TBUtility.get_values(datatype_config["value"], data, "json",
                                                  expression_instead_none=False)
                    values_tags = TBUtility.get_values(datatype_config["value"], data, "json",
                                                       get_tag=True)

                    keys = TBUtility.get_values(datatype_config["key"], data, "json",
                                                expression_instead_none=False)
                    keys_tags = TBUtility.get_values(datatype_config["key"], data, get_tag=True)

                    full_key = datatype_config["key"]
                    for (key, key_tag) in zip(keys, keys_tags):
                        is_valid_key = "${" in datatype_config["key"] and "}" in \
                                       datatype_config["key"]
                        full_key = full_key.replace('${' + str(key_tag) + '}',
                                                    str(key)) if is_valid_key else key_tag

                    full_value = datatype_config["value"]
                    for (value, value_tag) in zip(values, values_tags):
                        is_valid_value = "${" in datatype_config["value"] and "}" in \
                                         datatype_config["value"]

                        full_value = full_value.replace('${' + str(value_tag) + '}',
                                                        str(value)) if is_valid_value else value

                    if full_key != 'None' and full_value != 'None':
                        if datatype == 'timeseries' and (
                                data.get("ts") is not None or data.get("timestamp") is not None):
                            dict_result[self._datatypes[datatype]].append(
                                {"ts": data.get('ts', data.get('timestamp', int(time()))),
                                 'values': {full_key: full_value}})
                        else:
                            dict_result[self._datatypes[datatype]].append({full_key: full_value})

            return dict_result
        except json.decoder.JSONDecodeError:
            return None

    def _get_value(self, data, config, key):
        expression_arr = findall(r'\[\S[\d:]*]', config[key])
        if expression_arr:
            indexes = expression_arr[0][1:-1].split(':')
            try:
                if len(indexes) == 2:
                    from_index, to_index = indexes
                    return data[int(from_index) if from_index != '' else None:int(to_index) if to_index != '' else None]
                else:
                    index = int(indexes[0])
                    return data[index]
            except IndexError:
                self._log.error('deviceName expression invalid (index out of range)')
                return None
        else:
            return config[key]

    def _convert_text(self, val):
        dict_result = {'deviceName': self._get_value(val, self.__config, 'deviceNameExpression'),
                       'deviceType': self._get_value(val, self.__config, 'deviceTypeExpression')}

        if dict_result.get('deviceName') and dict_result.get('deviceType'):
            for datatype in self._datatypes:
                dict_result[self._datatypes[datatype]] = []
                for datatype_config in self.__config.get(datatype, []):
                    key = self._get_value(val, datatype_config, 'key')
                    value = self._get_value(val, datatype_config, 'value')

                    if key and value and datatype == 'timeseries':
                        dict_result[self._datatypes[datatype]].append({'ts': int(time()) * 1000, 'values': {key: value}})
                    else:
                        dict_result[self._datatypes[datatype]].append({key: value})

            return dict_result

        return None

    @StatisticsService.CollectStatistics(start_stat_type='receivedBytesFromDevices',
                                         end_stat_type='convertedBytesFromDevice')
    def convert(self, config, val):
        # convert data if it is json format
        result = self._convert_json(val)
        if result:
            return result

        # convert data using slices if it is text format
        result = self._convert_text(val)
        if result:
            return result

        # if none of above
        return None



#     Copyright 2022. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from simplejson import dumps
from time import time

from thingsboard_gateway.connectors.ocpp.ocpp_converter import OcppConverter
from thingsboard_gateway.tb_utility.tb_utility import TBUtility


class OcppUplinkConverter(OcppConverter):
    def __init__(self, config, logger):
        self._log = logger
        self.__config = config

    def get_device_name(self, config):
        try:
            if self.__config.get("deviceNameExpression") is not None:
                device_name_tags = TBUtility.get_values(self.__config.get("deviceNameExpression"), config,
                                                        get_tag=True)
                device_name_values = TBUtility.get_values(self.__config.get("deviceNameExpression"), config,
                                                          expression_instead_none=True)

                device_name = self.__config.get("deviceNameExpression")
                for (device_name_tag, device_name_value) in zip(device_name_tags, device_name_values):
                    is_valid_key = "${" in self.__config.get("deviceNameExpression") and "}" in \
                                   self.__config.get("deviceNameExpression")
                    device_name = device_name.replace('${' + str(device_name_tag) + '}',
                                                      str(device_name_value)) \
                        if is_valid_key else device_name_tag

                return device_name

            else:
                self._log.error("The expression for looking \"deviceName\" not found in config %s", dumps(self.__config))

        except Exception as e:
            self._log.error('Error in converter, for config: \n%s\n and message: \n%s\n %s', dumps(self.__config),
                            config, e)

    def get_device_type(self, config):
        try:
            if self.__config.get("deviceTypeExpression") is not None:
                device_type_tags = TBUtility.get_values(self.__config.get("deviceTypeExpression"), config,
                                                        get_tag=True)
                device_type_values = TBUtility.get_values(self.__config.get("deviceTypeExpression"), config,
                                                          expression_instead_none=True)
                device_type = self.__config.get("deviceTypeExpression")

                for (device_type_tag, device_type_value) in zip(device_type_tags, device_type_values):
                    is_valid_key = "${" in self.__config.get("deviceTypeExpression") and "}" in \
                                   self.__config.get("deviceTypeExpression")
                    device_type = device_type.replace('${' + str(device_type_tag) + '}',
                                                      str(device_type_value)) \
                        if is_valid_key else device_type_tag

                return device_type
        except Exception as e:
            self._log.error('Error in converter, for config: \n%s\n and message: \n%s\n %s', dumps(self.__config),
                            config, e)

    def convert(self, config, data):
        datatypes = {"attributes": "attributes",
                     "timeseries": "telemetry"}
        dict_result = {"deviceName": config['deviceName'], "deviceType": config['deviceType'], "attributes": [],
                       "telemetry": []}

        try:
            for datatype in datatypes:
                dict_result[datatypes[datatype]] = []

                for datatype_config in self.__config.get(datatype, []):
                    if config['messageType'] in datatype_config['messageTypeFilter'].split(','):
                        values = TBUtility.get_values(datatype_config["value"], data,
                                                      expression_instead_none=True)
                        values_tags = TBUtility.get_values(datatype_config["value"], data,
                                                           get_tag=True)

                        keys = TBUtility.get_values(datatype_config["key"], data,
                                                    expression_instead_none=True)
                        keys_tags = TBUtility.get_values(datatype_config["key"], data, get_tag=True)

                        full_key = datatype_config["key"]
                        for (key, key_tag) in zip(keys, keys_tags):
                            is_valid_key = "${" in datatype_config["key"] and "}" in \
                                           datatype_config["key"]
                            full_key = full_key.replace('${' + str(key_tag) + '}',
                                                        str(key)) if is_valid_key else key_tag

                        full_value = datatype_config["value"]
                        for (value, value_tag) in zip(values, values_tags):
                            is_valid_value = "${" in datatype_config["value"] and "}" in \
                                             datatype_config["value"]

                            full_value = full_value.replace('${' + str(value_tag) + '}',
                                                            str(value)) if is_valid_value else str(value)

                        if datatype == 'timeseries' and (
                                data.get("ts") is not None or data.get("timestamp") is not None):
                            dict_result[datatypes[datatype]].append(
                                {"ts": data.get('ts', data.get('timestamp', int(time()))),
                                 'values': {full_key: full_value}})
                        else:
                            dict_result[datatypes[datatype]].append({full_key: full_value})
        except Exception as e:
            self._log.error('Error in converter, for config: \n%s\n and message: \n%s\n %s', dumps(self.__config),
                            str(data), e)

        self._log.debug(dict_result)
        return dict_result



#     Copyright 2022. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

class StorageSettings:
    def __init__(self, config):
        self.data_folder_path = config.get("data_file_path", "./")
        self.messages_ttl_check_in_hours = config.get('messages_ttl_check_in_hours', 1) * 3600
        self.messages_ttl_in_days = config.get('messages_ttl_in_days', 7)



from os.path import exists, splitext
from os import path
from re import IGNORECASE, compile
from json import load, dump

from questionary import Validator, ValidationError, prompt
from pyfiglet import Figlet
from termcolor import colored
from yaml import safe_load


if exists('thingsboard_gateway/config/tb_gateway.json'):
    CONFIG_PATH = 'thingsboard_gateway/config/tb_gateway.json'
elif exists(path.dirname(path.abspath(__file__)) + '/config/tb_gateway.json'.replace('/', path.sep)):
    CONFIG_PATH = path.dirname(path.abspath(__file__)) + '/config/tb_gateway.json'.replace('/', path.sep)
elif exists("/etc/thingsboard-gateway/config/tb_gateway.json".replace('/', path.sep)):
    CONFIG_PATH = "/etc/thingsboard-gateway/config/tb_gateway.json".replace('/', path.sep)
elif exists('thingsboard_gateway/config/tb_gateway.yaml'):
    CONFIG_PATH = 'thingsboard_gateway/config/tb_gateway.yaml'
elif exists(path.dirname(path.abspath(__file__)) + '/config/tb_gateway.yaml'.replace('/', path.sep)):
    CONFIG_PATH = path.dirname(path.abspath(__file__)) + '/config/tb_gateway.yaml'.replace('/', path.sep)
elif exists("/etc/thingsboard-gateway/config/tb_gateway.yaml".replace('/', path.sep)):
    CONFIG_PATH = "/etc/thingsboard-gateway/config/tb_gateway.yaml".replace('/', path.sep)

class NotNullValidator(Validator):
    def validate(self, document):
        if not document.text or document.text == '':
            raise ValidationError(message='Value can be empty!', cursor_position=len(document.text))


class NumberValidator(Validator):
    def validate(self, document):
        try:
            int(document.text)
        except ValueError:
            raise ValidationError(message='Must be a number type!', cursor_position=len(document.text))


class PortValidator(Validator):
    def validate(self, document):
        try:
            port = int(document.text)
        except ValueError:
            raise ValidationError(message='Must be a number type!', cursor_position=len(document.text))

        if not 1 <= port <= 65535:
            raise ValidationError(message='Port is invalid!', cursor_position=len(document.text))


class HostValidator(NotNullValidator):
    def validate(self, document):
        super(HostValidator, self).validate(document)

        hostname = document.text
        if len(hostname) > 255:
            return ValidationError(message='Host is invalid!', cursor_position=len(document.text))

        if hostname[-1] == ".":
            hostname = hostname[:-1]

        allowed = compile(r"(?!-)[A-Z\d-]{1,63}(?<!-)$", IGNORECASE)
        if not all(allowed.match(x) for x in hostname.split(".")):
            raise ValidationError(message='Host is invalid!', cursor_position=len(document.text))


class PathValidator(Validator):
    def validate(self, document):
        if not exists(document.text):
            raise ValidationError(message='File doesn\'t exist!', cursor_position=len(document.text))

        if splitext(document.text)[1] != '.pem':
            raise ValidationError(message='File must be .pem extension!', cursor_position=len(document.text))


class FileExtensionValidator(NotNullValidator):
    def validate(self, document):
        super(FileExtensionValidator, self).validate(document)

        if document.text.split('.')[-1] != 'json':
            raise ValidationError(message='File must be .json!', cursor_position=len(document.text))


def read_config_file() -> {str: str}:
    file_extension = CONFIG_PATH.split('.')[-1]
    if file_extension == 'json':
        try:
            with open(CONFIG_PATH, 'r') as file:
                config_dict = load(file)
        except Exception as e:
            print('Failed to load configuration file:\n %s', e)
    else:
        print('YAML configuration is deprecated. '
                    'Please, use JSON configuration instead.')
        print(
            'See default configuration on '
            'https://thingsboard.io/docs/iot-gateway/configuration/?storageConfig=sqlite#storage-configuration')
        try:
            with open(CONFIG_PATH, 'r') as general_config:
                config_dict = safe_load(general_config)
        except Exception as e:
            print('Failed to load configuration file:\n %s', e)

    return config_dict


def generate_config_file(data: {str: str}) -> None:
    config_path = CONFIG_PATH
    if CONFIG_PATH.split('.')[-1] == 'yaml':
        config_path = ''.join(CONFIG_PATH.split('.')[:-1]) + '.json'

    with open(config_path, 'w') as file:
        dump(data, file, sort_keys=False, indent='  ')


def configure():
    try:
        default_config = read_config_file()
        connectors_list = []
        grpc_enabled = False
        base_answers = default_config['thingsboard']
        statistics_answers = default_config['thingsboard'].get('statistics', {})
        grpc_api_answers = default_config.get('grpc', {})
        storage_answers = default_config.get('storage', {})
        qos_and_storage_type_answers = {
            'qos': 1,
            'storage': 'memory'
        }

        # GENERAL SETTINGS ---------------------------------------------------------------------------------------------
        simple_setup_questions = [
            {
                'type': 'input',
                'name': 'host',
                'message': 'ThingsBoard host:',
                'default': default_config['thingsboard']['host'],
                'validate': HostValidator
            },
            {
                'type': 'input',
                'name': 'port',
                'message': 'ThingsBoard port:',
                'default': str(default_config['thingsboard']['port']),
                'validate': PortValidator,
                'filter': lambda val: int(val)
            },
            {
                'type': 'confirm',
                'name': 'remoteConfiguration',
                'message': 'Do you want to enable remote configuration feature? (No)',
                'default': True
            },
            {
                'type': 'list',
                'name': 'security',
                'message': 'What security type do you need?',
                'choices': [
                    'Access Token (Basic Security)',
                    'Username and Password (Basic Security)',
                    'TLS + Access Token (Advanced Security)',
                    'TLS + Private Key (Advanced Security)'
                ]
            }
        ]

        base_questions = [
            {
                'type': 'confirm',
                'name': 'remoteShell',
                'message': 'Do you want to have access from remote shell? (No)',
                'default': False
            },
            {
                'type': 'input',
                'name': 'minPackSendDelayMS',
                'message': 'Minimal delay between sending messages (milliseconds):',
                'default': str(default_config['thingsboard']['minPackSendDelayMS']),
                'validate': NumberValidator,
                'filter': lambda val: int(val)
            },
            {
                'type': 'input',
                'name': 'maxPayloadSizeBytes',
                'message': 'Maximum payload size (bytes):',
                'default': str(default_config['thingsboard']['maxPayloadSizeBytes']),
                'validate': NumberValidator,
                'filter': lambda val: int(val)
            },
            {
                'type': 'input',
                'name': 'minPackSizeToSend',
                'message': 'Minimal pack size for sending (bytes):',
                'default': str(default_config['thingsboard']['minPackSizeToSend']),
                'validate': NumberValidator,
                'filter': lambda val: int(val)
            },
            {
                'type': 'confirm',
                'name': 'handleDeviceRenaming',
                'message': 'Handle device renaming?',
                'default': bool(default_config['thingsboard']['handleDeviceRenaming']),
            },
            {
                'type': 'input',
                'name': 'checkConnectorsConfigurationInSeconds',
                'message': 'Connectors config files will check every (sec.):',
                'default': str(default_config['thingsboard']['checkConnectorsConfigurationInSeconds']),
                'validate': NumberValidator,
                'filter': lambda val: int(val)
            },
            {
                'type': 'confirm',
                'name': 'grpc-enabled',
                'message': 'Do you want to enable GRPC API on your gateway?',
                'default': False
            },
            {
                'type': 'confirm',
                'name': 'statistics-enabled',
                'message': 'Do you want to enable statistics collecting?',
                'default': True
            }
        ]

        f = Figlet(font='slant')
        print(colored(f.renderText('ThingsBoard'), color='white'))
        print(colored(f.renderText('IoT Gateway'), color='red'))
        print(colored('Welcome to ThingsBoard IoT Gateway configuration Wizard', 'cyan'))
        print(colored('Let\'s configure you Gateway by answering on questions below ⬇\n'))

        simple_setup_answers = prompt(simple_setup_questions)

        # SECURITY SETTINGS --------------------------------------------------------------------------------------------
        access_token_config = [
            {
                'type': 'input',
                'name': 'accessToken',
                'message': 'Your token:',
                'validate': NotNullValidator
            }
        ]
        tls = [
            {
                'type': 'input',
                'name': 'caCert',
                'message': 'Path to your CA file (.pem):',
                'validate': PathValidator
            }
        ]
        basic_config = [
            {
                'type': 'input',
                'name': 'username',
                'message': 'Your username:'
            },
            {
                'type': 'input',
                'name': 'password',
                'message': 'Your password:'
            },
            {
                'type': 'input',
                'name': 'clientId',
                'message': 'Your client ID:'
            }
        ]
        tls_access_token_config = access_token_config + tls
        tls_private_key_config = [
                                     {
                                         'type': 'input',
                                         'name': 'privateKey',
                                         'message': 'Path to you private key file (.pem):',
                                         'validate': PathValidator
                                     }
                                 ] + tls + [
                                     {
                                         'type': 'input',
                                         'name': 'cert',
                                         'message': 'Path to your certificate file (.pem):',
                                         'validate': PathValidator
                                     }
                                 ]

        if simple_setup_answers['security'] == 'Access Token (Basic Security)':
            security_questions = access_token_config
        elif simple_setup_answers['security'] == 'TLS + Access Token (Advanced Security)':
            security_questions = tls_access_token_config
        elif simple_setup_answers['security'] == 'Username and Password (Basic Security)':
            security_questions = basic_config
        else:
            security_questions = tls_private_key_config

        security_answers = prompt(security_questions)

        continue_answer = prompt([
            {
                'type': 'confirm',
                'name': 'continue',
                'message': 'Continue to advance settings(Y) or save and exit(N)?',
                'default': False
            }
        ])

        if continue_answer['continue']:
            base_answers = prompt(base_questions)

            # STATISTICS SETTINGS --------------------------------------------------------------------------------------
            statistics_answers = {}
            if base_answers.pop('statistics-enabled'):
                statistics_config = [
                    {
                        'type': 'input',
                        'name': 'statsSendPeriodInSeconds',
                        'message': 'Period of time sending statistics (sec.):',
                        'default': str(default_config['thingsboard']['statistics']['statsSendPeriodInSeconds']),
                        'validate': NumberValidator,
                        'filter': lambda val: int(val)
                    },
                    {
                        'type': 'input',
                        'name': 'configuration',
                        'message': 'Config file of custom statistic commands:'
                    }
                ]

                statistics_answers = prompt(statistics_config)
                statistics_answers['enable'] = True

            # GRPC SETTINGS ------------------------------------------------------------------------------------------------
            grpc_enabled = base_answers.pop('grpc-enabled')

            if grpc_enabled:
                grpc_api_questions = [
                    {
                        'name': 'serverPort',
                        'default': 9595,
                        'message': '[GRPC] Please set port for GRPC server:',
                        'validate': NumberValidator,
                        'filter': lambda val: int(val)
                    },
                    {
                        'name': 'keepaliveTimeMs',
                        'default': 10000,
                        'message': '[GRPC] Keep alive period:',
                        'validate': NumberValidator,
                        'filter': lambda val: int(val)
                    },
                    {
                        'name': 'keepaliveTimeoutMs',
                        'default': 5000,
                        'message': '[GRPC] Keep alive timeout',
                        'validate': NumberValidator,
                        'filter': lambda val: int(val)
                    },
                    {
                        'name': 'keepalivePermitWithoutCalls',
                        'default': True,
                        'message': '[GRPC] Allow send pings from clients without calls:'
                    },
                    {
                        'name': 'maxPingsWithoutData',
                        'default': 0,
                        'message': '[GRPC] Maximal count of pings without data from client to server:',
                        'validate': NumberValidator,
                        'filter': lambda val: int(val)
                    },
                    {
                        'name': 'minTimeBetweenPingsMs',
                        'default': 10000,
                        'message': '[GRPC] Minimal period between ping messages:',
                        'validate': NumberValidator,
                        'filter': lambda val: int(val)
                    },
                    {
                        'name': 'minPingIntervalWithoutDataMs',
                        'default': 5000,
                        'message': '[GRPC] Minimal period between ping messages without data:',
                        'validate': NumberValidator,
                        'filter': lambda val: int(val)
                    },
                ]
                grpc_api_answers = prompt(grpc_api_questions)
            else:
                grpc_api_answers = default_config.get('grpc', {})

            # QOS AND STORAGE TYPE SETTINGS ----------------------------------------------------------------------------
            qos_and_storage_type_question = [
                {
                    'type': 'input',
                    'name': 'qos',
                    'message': 'QoS:',
                    'validate': NumberValidator,
                    'default': str(default_config['thingsboard']['qos']),
                    'filter': lambda val: int(val)
                },
                {
                    'type': 'list',
                    'name': 'storage',
                    'message': 'Choose storage type:',
                    'choices': [
                        'Memory',
                        'File storage',
                        'SQLite'
                    ],
                    'filter': lambda val: 'file' if val == 'File storage' else val.lower()
                }
            ]

            qos_and_storage_type_answers = prompt(qos_and_storage_type_question)

            # STORAGE SETTINGS -----------------------------------------------------------------------------------------
            if qos_and_storage_type_answers['storage'] == 'memory':
                storage_questions = [
                    {
                        'type': 'input',
                        'name': 'read_records_count',
                        'message': 'Count of messages to get from storage and send to ThingsBoard:',
                        'default': str(default_config['storage'].get('read_records_count', '')),
                        'validate': NumberValidator,
                        'filter': lambda val: int(val)
                    },
                    {
                        'type': 'input',
                        'name': 'max_records_count',
                        'message': 'Maximum count of data in storage before send to ThingsBoard:',
                        'default': str(default_config['storage'].get('max_records_count', '')),
                        'validate': NumberValidator,
                        'filter': lambda val: int(val)
                    }
                ]
            elif qos_and_storage_type_answers['storage'] == 'file':
                storage_questions = [
                    {
                        'type': 'input',
                        'name': 'data_folder_path',
                        'message': 'Path to folder, that will contains data (Relative or Absolute):',
                        'default': str(default_config['storage'].get('data_folder_path', '')),
                        'validate': NotNullValidator
                    },
                    {
                        'type': 'input',
                        'name': 'max_file_count',
                        'message': 'Maximum count of file that will be saved:',
                        'default': str(default_config['storage'].get('max_file_count', '')),
                        'validate': NumberValidator,
                        'filter': lambda val: int(val)
                    },
                    {
                        'type': 'input',
                        'name': 'max_read_records_count',
                        'message': 'Count of messages to get from storage and send to ThingsBoard:',
                        'default': str(default_config['storage'].get('max_read_records_count', '')),
                        'validate': NumberValidator,
                        'filter': lambda val: int(val)
                    },
                    {
                        'type': 'input',
                        'name': 'max_records_per_file',
                        'message': 'Maximum count of records that will be stored in one file:',
                        'default': str(default_config['storage'].get('max_records_per_file', '')),
                        'validate': NumberValidator,
                        'filter': lambda val: int(val)
                    }
                ]
            else:
                storage_questions = [
                    {
                        'type': 'input',
                        'name': 'data_file_path',
                        'message': 'Path to folder, that will contains data (Relative or Absolute):',
                        'default': str(default_config['storage'].get('data_file_path', '')),
                        'validate': NotNullValidator
                    },
                    {
                        'type': 'input',
                        'name': 'messages_ttl_check_in_hours',
                        'message': 'How often will Gateway check data for obsolescence:',
                        'default': str(default_config['storage'].get('messages_ttl_check_in_hours', '')),
                        'validate': NumberValidator,
                        'filter': lambda val: int(val)
                    },
                    {
                        'type': 'input',
                        'name': 'messages_ttl_in_days',
                        'message': 'Maximum days that storage will save data:',
                        'default': str(default_config['storage'].get('messages_ttl_in_days', '')),
                        'validate': NumberValidator,
                        'filter': lambda val: int(val)
                    }
                ]

            storage_answers = prompt(storage_questions)

            # CONNECTORS SETTINGS --------------------------------------------------------------------------------------
            connectors_questions = [
                {
                    'type': 'checkbox',
                    'name': 'connectors',
                    'message': 'Choose connectors you want to use:',
                    'choices': [
                        {
                            'name': 'MQTT',
                        },
                        {
                            'name': 'FTP',
                        },
                        {
                            'name': 'Modbus',
                        },
                        {
                            'name': 'CAN',
                        },
                        {
                            'name': 'Bacnet',
                        },
                        {
                            'name': 'BLE',
                        },
                        {
                            'name': 'OPC-UA',
                        },
                        {
                            'name': 'OPC-UA AsyncIO'
                        },
                        {
                            'name': 'ODBC',
                        },
                        {
                            'name': 'Request',
                        },
                        {
                            'name': 'REST',
                        },
                        {
                            'name': 'SNMP'
                        },
                        {
                            'name': 'XMPP'
                        },
                        {
                            'name': 'OCPP'
                        },
                        {
                            'name': 'Socket'
                        }
                    ],
                    'validate': lambda answer: 'You must choose at least one connector.' if len(answer) == 0 else True
                }
            ]

            connectors_answers = prompt(connectors_questions)

            for connector in connectors_answers['connectors']:
                print(colored(f'Configuration {connector} connector:', 'blue'))
                connector_questions = [
                    {
                        'type': 'input',
                        'name': 'name',
                        'message': 'Name of connector:',
                        'validate': NotNullValidator
                    },
                    {
                        'type': 'input',
                        'name': 'configuration',
                        'message': 'Config file of connector:',
                        'validate': FileExtensionValidator
                    }
                ]
                connector_answers = prompt(connector_questions)
                connectors_list.append({'type': connector.lower(), **connector_answers})

        generate_config_file(
            {
                'thingsboard': {**base_answers, **simple_setup_answers, 'security': security_answers,
                                'qos': qos_and_storage_type_answers['qos'], 'statistics': statistics_answers},
                'storage': {'type': qos_and_storage_type_answers['storage'], **storage_answers},
                'grpc': {'enabled': grpc_enabled, **grpc_api_answers},
                'connectors': connectors_list
            })

        print(colored('Configuration file updated 🎉\nDon\'t forget to restart your gateway to apply changes!'))
    except Exception as e:
        print(colored('Something went wrong! Please try again.', color='red'))
        raise e


if __name__ == '__main__':
    configure()



#     Copyright 2022. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import socket
from queue import Queue
from random import choice
from re import findall
from string import ascii_lowercase
from threading import Thread
from time import sleep

from simplejson import dumps

from thingsboard_gateway.connectors.connector import Connector
from thingsboard_gateway.tb_utility.tb_loader import TBModuleLoader
from thingsboard_gateway.gateway.statistics_service import StatisticsService
from thingsboard_gateway.connectors.socket.socket_decorators import CustomCollectStatistics
from thingsboard_gateway.tb_utility.tb_logger import init_logger

SOCKET_TYPE = {
    'TCP': socket.SOCK_STREAM,
    'UDP': socket.SOCK_DGRAM
}
DEFAULT_UPLINK_CONVERTER = 'BytesSocketUplinkConverter'


class SocketConnector(Connector, Thread):
    def __init__(self, gateway, config, connector_type):
        super().__init__()
        self.__config = config
        self._connector_type = connector_type
        self.statistics = {'MessagesReceived': 0,
                           'MessagesSent': 0}
        self.__gateway = gateway
        self.setName(config.get("name", 'TCP Connector ' + ''.join(choice(ascii_lowercase) for _ in range(5))))
        self.__log = init_logger(self.__gateway, self.name, self.__config.get('logLevel', 'INFO'))
        self.daemon = True
        self.__stopped = False
        self._connected = False
        self.__bind = False
        self.__socket_type = config['type'].upper()
        self.__socket_address = config['address']
        self.__socket_port = config['port']
        self.__socket_buff_size = config['bufferSize']
        self.__socket = socket.socket(socket.AF_INET, SOCKET_TYPE[self.__socket_type])
        self.__converting_requests = Queue(-1)

        self.__devices = self.__convert_devices_list()
        self.__connections = {}

    def __convert_devices_list(self):
        devices = self.__config.get('devices', [])

        converted_devices = {}
        for device in devices:
            address = device.get('address')
            module = self.__load_converter(device)
            converter = module(
                {'deviceName': device['deviceName'],
                 'deviceType': device.get('deviceType', 'default')}, self.__log) if module else None
            device['converter'] = converter

            # validate attributeRequests requestExpression
            attr_requests = device.get('attributeRequests', [])
            device['attributeRequests'] = self.__validate_attr_requests(attr_requests)
            if len(device['attributeRequests']):
                self.__attribute_type = {
                    'client': self.__gateway.tb_client.client.gw_request_client_attributes,
                    'shared': self.__gateway.tb_client.client.gw_request_shared_attributes
                }

            converted_devices[address] = device

        return converted_devices

    def __load_converter(self, device):
        converter_class_name = device.get('converter', DEFAULT_UPLINK_CONVERTER)
        module = TBModuleLoader.import_module(self._connector_type, converter_class_name)

        if module:
            self.__log.debug('Converter %s for device %s - found!', converter_class_name, self.name)
            return module

        self.__log.error("Cannot find converter for %s device", self.name)
        return None

    def __validate_attr_requests(self, attr_requests):
        validated_attrs = []
        for attr in attr_requests:
            valid_attr = False
            if attr['requestExpression'] != '':
                if '${' in attr['requestExpression'] and '}' in attr['requestExpression']:
                    if '==' in attr['requestExpression']:
                        expression_arr = findall(r'\[[^\s][0-9:]*]', attr['requestExpression'])
                        if expression_arr:
                            indexes = expression_arr[0][1:-1].split(':')
                            if len(indexes) == 2:
                                from_index, to_index = indexes
                                attr['requestIndexFrom'] = from_index
                                attr['requestIndexTo'] = to_index
                            else:
                                attr['requestIndex'] = int(indexes[0])

                            attr['haveIndex'] = True
                            try:
                                attr['requestEqual'] = attr['requestExpression'].split('==')[-1][:-1]
                            except IndexError:
                                self.__log.error(f'{attr["requestExpression"]} not valid. Index out of range.')
                                continue

                            valid_attr = True
                            validated_attrs.append(attr)
                else:
                    valid_attr = True
                    attr['haveIndex'] = False
                    validated_attrs.append(attr)

            if not valid_attr:
                self.__log.error(f'{attr["requestExpression"]} not valid expression')

        return validated_attrs

    def open(self):
        self.__stopped = False
        self.start()

    def run(self):
        self._connected = True

        converting_thread = Thread(target=self.__process_data, daemon=True, name='Converter Thread')
        converting_thread.start()

        while not self.__bind:
            try:
                self.__socket.bind((self.__socket_address, self.__socket_port))
            except OSError:
                self.__log.error('Address already in use. Reconnecting...')
                sleep(3)
            else:
                self.__bind = True

        if self.__socket_type == 'TCP':
            self.__socket.listen(5)

        self.__log.info('%s socket is up', self.__socket_type)

        while not self.__stopped:
            try:
                if self.__socket_type == 'TCP':
                    conn, address = self.__socket.accept()
                    self.__connections[address] = conn

                    self.__log.debug('New connection %s established', address)
                    thread = Thread(target=self.__process_tcp_connection, daemon=True,
                                    name=f'Processing {address} connection',
                                    args=(conn, address))
                    thread.start()
                else:
                    data, client_address = self.__socket.recvfrom(self.__socket_buff_size)
                    self.__converting_requests.put((client_address, data))
            except ConnectionAbortedError:
                self.__socket.close()

    def __process_tcp_connection(self, connection, address):
        while not self.__stopped:
            data = connection.recv(self.__socket_buff_size)

            if data:
                self.__converting_requests.put((address, data))
            else:
                break

        connection.close()
        self.__connections.pop(address)
        self.__log.debug('Connection %s closed', address)

    def __process_data(self):
        while not self.__stopped:
            if not self.__converting_requests.empty():
                (address, port), data = self.__converting_requests.get()

                device = self.__devices.get(f'{address}:{port}', None)
                if not device:
                    self.__log.error('Can\'t convert data from %s:%s - not in config file', address, port)

                # check data for attribute requests
                is_attribute_request = False
                attr_requests = device.get('attributeRequests', [])
                if len(attr_requests):
                    for attr in attr_requests:
                        equal = data
                        if attr['haveIndex']:
                            if attr.get('requestIndexFrom') and attr.get('requestIndexTo'):
                                index_from = int(attr['requestIndexFrom']) if attr['requestIndexFrom'] != '' else None
                                index_to = int(attr['requestIndexTo']) if attr['requestIndexTo'] != '' else None
                                equal = data[index_from:index_to]
                            else:
                                equal = data[int(attr['requestIndex'])]

                        if attr['requestEqual'] == equal.decode('utf-8'):
                            is_attribute_request = True
                            self.__process_attribute_request(device['deviceName'], attr, data)

                    if is_attribute_request:
                        continue

                self.__convert_data(device, data)

            sleep(.2)

    def __convert_data(self, device, data):
        address, port = device['address'].split(':')
        converter = device['converter']
        if not converter:
            self.__log.error('Converter not found for %s:%s', address, port)

        try:
            device_config = {
                'encoding': device.get('encoding', 'utf-8').lower(),
                'telemetry': device.get('telemetry', []),
                'attributes': device.get('attributes', [])
            }
            converted_data = converter.convert(device_config, data)

            self.statistics['MessagesReceived'] = self.statistics['MessagesReceived'] + 1

            if converted_data is not None:
                self.__gateway.send_to_storage(self.get_name(), converted_data)
                self.statistics['MessagesSent'] = self.statistics['MessagesSent'] + 1
                self.__log.info('Data to ThingsBoard %s', converted_data)
        except Exception as e:
            self.__log.exception(e)

    def __process_attribute_request(self, device_name, attr, data):
        expression_arr = findall(r'\[[^\s][0-9:]*]', attr['attributeNameExpression'])

        found_attributes = []
        if expression_arr:
            for exp in expression_arr:
                indexes = exp[1:-1].split(':')

                try:
                    if len(indexes) == 2:
                        from_index, to_index = indexes
                        attribute = data[int(from_index) if from_index != '' else None:int(
                            to_index) if to_index != '' else None]
                    else:
                        attribute = data[int(indexes[0])]

                    found_attributes.append(attribute.decode('utf-8'))
                except IndexError:
                    self.__log.error('Data length not valid due to attributeNameExpression')
                    return

        self.statistics['MessagesReceived'] = self.statistics['MessagesReceived'] + 1
        self.__attribute_type[attr['type']](device_name, found_attributes, self.__attribute_request_callback)

    def __attribute_request_callback(self, response, _):
        device = response.get('device')
        if not device:
            self.__log.error('Attribute request does\'t return device name')

        device = tuple(filter(lambda item: item['deviceName'] == device, self.__config['devices']))[0]
        address, port = device['address'].split(':')

        value = response.get('value') or response.get('values')
        converted_value = bytes(dumps(value), encoding='utf-8')

        if self.__socket_type == 'TCP':
            self.__write_value_via_tcp(address, port, converted_value)
        else:
            self.__write_value_via_udp(address, port, converted_value)
        self.statistics['MessagesSent'] = self.statistics['MessagesSent'] + 1

    def close(self):
        self.__stopped = True
        self._connected = False
        self.__connections = {}
        self.__log.reset()

    def get_name(self):
        return self.name

    def get_type(self):
        return self._connector_type

    def is_connected(self):
        return self._connected

    def get_config(self):
        return self.__config

    @CustomCollectStatistics(start_stat_type='allBytesSentToDevices')
    def __write_value_via_tcp(self, address, port, value):
        try:
            self.__connections[(address, int(port))].sendall(value)
            return 'ok'
        except KeyError:
            try:
                new_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                new_socket.connect((address, int(port)))
                new_socket.sendall(value)
                new_socket.close()
                return 'ok'
            except ConnectionRefusedError as e:
                self.__log.error('Can\'t connect to %s:%s\n %s', address, port, e)
                return e

    @staticmethod
    @StatisticsService.CollectStatistics(start_stat_type='allBytesSentToDevices')
    def __write_value_via_udp(address, port, value):
        new_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        new_socket.sendto(value, (address, int(port)))
        new_socket.close()

    @StatisticsService.CollectAllReceivedBytesStatistics(start_stat_type='allReceivedBytesFromTB')
    def on_attributes_update(self, content):
        try:
            device = tuple(filter(lambda item: item['deviceName'] == content['device'], self.__config['devices']))[0]

            for attribute_update_config in device['attributeUpdates']:
                for attribute_update in content['data']:
                    if attribute_update_config['attributeOnThingsBoard'] == attribute_update:
                        address, port = device['address'].split(':')
                        encoding = device.get('encoding', 'utf-8').lower()
                        converted_data = bytes(str(content['data'][attribute_update]), encoding=encoding)
                        self.__write_value_via_tcp(address, port, converted_data)
        except IndexError:
            self.__log.error('Device not found')

    @StatisticsService.CollectAllReceivedBytesStatistics(start_stat_type='allReceivedBytesFromTB')
    def server_side_rpc_handler(self, content):
        try:
            if content.get('data') is None:
                content['data'] = {'params': content['params'], 'method': content['method']}

            rpc_method = content['data']['method']

            # check if RPC type is connector RPC (can be only 'set')
            try:
                (connector_type, rpc_method_name) = rpc_method.split('_')
                if connector_type == self._connector_type:
                    rpc_method = rpc_method_name
                    content['device'] = content['params'].split(' ')[0].split('=')[-1]
            except (IndexError, ValueError):
                pass

            device = tuple(filter(lambda item: item['deviceName'] == content['device'], self.__config['devices']))[0]

            # check if RPC method is reserved set
            if rpc_method == 'set':
                params = {}
                for param in content['data']['params'].split(';'):
                    try:
                        (key, value) = param.split('=')
                    except ValueError:
                        continue

                    if key and value:
                        params[key] = value

                result = None
                try:
                    if self.__socket_type == 'TCP':
                        result = self.__write_value_via_tcp(params['address'], int(params['port']), params['value'])
                    else:
                        self.__write_value_via_udp(params['address'], int(params['port']), params['value'])
                except KeyError:
                    self.__gateway.send_rpc_reply(device=device, req_id=content['data'].get('id'),
                                                  content='Not enough params')
                except ValueError:
                    self.__gateway.send_rpc_reply(device=device, req_id=content['data']['id'],
                                                  content='Param "port" have to be int type')
                else:
                    self.__gateway.send_rpc_reply(device=device, req_id=content['data'].get('id'), content=str(result))
            else:
                for rpc_config in device['serverSideRpc']:
                    for (key, value) in content['data'].items():
                        if value == rpc_config['methodRPC']:
                            rpc_method = rpc_config['methodProcessing']
                            return_result = rpc_config['withResponse']
                            result = None

                            address, port = device['address'].split(':')
                            encoding = device.get('encoding', 'utf-8').lower()
                            converted_data = bytes(str(content['data']['params']), encoding=encoding)

                            if rpc_method.upper() == 'WRITE':
                                if self.__socket_type == 'TCP':
                                    result = self.__write_value_via_tcp(address, port, converted_data)
                                else:
                                    self.__write_value_via_udp(address, port, converted_data)

                            if return_result and self.__socket_type == 'TCP':
                                self.__gateway.send_rpc_reply(content['device'], content['data']['id'], str(result))

                            return
        except IndexError:
            self.__log.error('Device not found')
        except Exception as e:
            self.__log.exception(e)



#      Copyright 2022. ThingsBoard
#  #
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#  #
#          http://www.apache.org/licenses/LICENSE-2.0
#  #
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

from thingsboard_gateway.connectors.converter import Converter, log
from thingsboard_gateway.gateway.proto.messages_pb2 import ConnectMsg, DisconnectMsg, GatewayAttributesMsg, GatewayAttributesRequestMsg, GatewayClaimMsg, \
    GatewayRpcResponseMsg, GatewayTelemetryMsg, KeyValueProto, KeyValueType, Response


class GrpcUplinkConverter(Converter):
    def __init__(self):
        self.__conversion_methods = {
            Response.DESCRIPTOR: self.__convert_response_msg,
            GatewayTelemetryMsg.DESCRIPTOR: self.__convert_gateway_telemetry_msg,
            GatewayAttributesMsg.DESCRIPTOR: self.__convert_gateway_attributes_msg,
            GatewayClaimMsg.DESCRIPTOR: self.__convert_gateway_claim_msg,
            ConnectMsg.DESCRIPTOR: self.__convert_connect_msg,
            DisconnectMsg.DESCRIPTOR: self.__convert_disconnect_msg,
            GatewayRpcResponseMsg.DESCRIPTOR: self.__convert_gateway_rpc_response_msg,
            GatewayAttributesRequestMsg.DESCRIPTOR: self.__convert_gateway_attributes_request_msg
            }

    def convert(self, config, data):
        try:
            if self.__conversion_methods.get(data.DESCRIPTOR) is not None:
                return self.__conversion_methods[data.DESCRIPTOR](data)
            else:
                log.error("[GRPC] unknown uplink message descriptor: %r", data.DESCRIPTOR)
                return {}
        except Exception as e:
            log.exception("[GRPC] ", e)
            return {}

    @staticmethod
    def __convert_response_msg(self, msg: Response):
        log.debug("Converted response: %r", msg.response)

    @staticmethod
    def __convert_gateway_telemetry_msg(msg: GatewayTelemetryMsg):
        result = []
        for telemetry_msg in msg.msg:
            device_dict = {"deviceName": telemetry_msg.deviceName, "telemetry": []}
            for ts_kv_list in telemetry_msg.msg.tsKvList:
                ts_kv_list_dict = {"ts": ts_kv_list.ts, "values": {}}
                for kv in ts_kv_list.kv:
                    ts_kv_list_dict['values'][kv.key] = GrpcUplinkConverter.get_value(kv)
                device_dict['telemetry'].append(ts_kv_list_dict)
            result.append(device_dict)
        return result

    @staticmethod
    def __convert_gateway_attributes_msg(msg: GatewayAttributesMsg):
        result = []
        for attributes_msg in msg.msg:
            device_dict = {"deviceName": attributes_msg.deviceName, "attributes": {}}
            for kv in attributes_msg.msg.kv:
                device_dict['attributes'][kv.key] = GrpcUplinkConverter.get_value(kv)
            result.append(device_dict)
        return result

    @staticmethod
    def __convert_gateway_claim_msg(msg: GatewayClaimMsg):
        result_dict = {}
        for claim_device_msg in msg:
            result_dict.update({claim_device_msg.deviceName: {"secretKey": claim_device_msg.msg.secretKey, "durationMs": claim_device_msg.msg.durationMs}})
        return result_dict

    @staticmethod
    def __convert_connect_msg(msg: ConnectMsg) -> dict:
        result_dict = {'deviceName': msg.deviceName, 'deviceType': msg.deviceType}
        return result_dict

    @staticmethod
    def __convert_disconnect_msg(msg: DisconnectMsg):
        return {"deviceName": msg.deviceName}

    @staticmethod
    def __convert_gateway_rpc_response_msg(msg: GatewayRpcResponseMsg):
        return {"deviceName": msg.deviceName, "id": msg.id, "data": msg.data}

    @staticmethod
    def __convert_gateway_attributes_request_msg(msg: GatewayAttributesRequestMsg):
        return {"id": msg.id, "deviceName": msg.deviceName, "client": msg.client, "keys": msg.keys}

    @staticmethod
    def get_value(msg: KeyValueProto):
        if msg.type == KeyValueType.BOOLEAN_V:
            return msg.bool_v
        if msg.type == KeyValueType.LONG_V:
            return msg.long_v
        if msg.type == KeyValueType.DOUBLE_V:
            return msg.double_v
        if msg.type == KeyValueType.STRING_V:
            return msg.string_v
        if msg.type == KeyValueType.JSON_V:
            return msg.json_v



import time
from re import findall

from simplejson import dumps

from thingsboard_gateway.connectors.mqtt.mqtt_uplink_converter import MqttUplinkConverter
from thingsboard_gateway.gateway.statistics_service import StatisticsService


class BytesMqttUplinkConverter(MqttUplinkConverter):
    def __init__(self, config, logger):
        self.__config = config.get('converter')
        self._log = logger

    @property
    def config(self):
        return self.__config

    @config.setter
    def config(self, value):
        self.__config = value

    @StatisticsService.CollectStatistics(start_stat_type='receivedBytesFromDevices',
                                         end_stat_type='convertedBytesFromDevice')
    def convert(self, topic, data):
        datatypes = {"attributes": "attributes",
                     "timeseries": "telemetry"}
        dict_result = {
            "deviceName": self.parse_data(self.__config['deviceNameExpression'], data),
            "deviceType": self.parse_data(self.__config['deviceTypeExpression'], data),
            "attributes": [],
            "telemetry": []
        }

        try:
            for datatype in datatypes:
                dict_result[datatypes[datatype]] = []
                for datatype_config in self.__config.get(datatype, []):
                    value_item = {datatype_config['key']: self.parse_data(datatype_config['value'], data)}
                    if datatype == 'timeseries':
                        dict_result[datatypes[datatype]].append({"ts": int(time.time()) * 1000, 'values': value_item})
                    else:
                        dict_result[datatypes[datatype]].append(value_item)
        except Exception as e:
            self._log.error('Error in converter, for config: \n%s\n and message: \n%s\n %s', dumps(self.__config),
                            str(data), e)

        self._log.debug('Converted data: %s', dict_result)
        return dict_result

    @staticmethod
    def parse_data(expression, data):
        expression_arr = findall(r'\[\S[0-9:]*]', expression)
        converted_data = expression

        for exp in expression_arr:
            indexes = exp[1:-1].split(':')

            data_to_replace = ''
            if len(indexes) == 2:
                from_index, to_index = indexes
                concat_arr = data[
                             int(from_index) if from_index != '' else None:int(
                                 to_index) if to_index != '' else None]
                for sub_item in concat_arr:
                    data_to_replace += str(sub_item)
            else:
                data_to_replace += str(data[int(indexes[0])])

            converted_data = converted_data.replace(exp, data_to_replace)

        return converted_data
